
# interfaces the CBD modelling&simulation environment
# with the "generic" plotter (in the central Kernel/Plotting/ directory)

from CBD_Constants import *

#wrapper for class methods.            
class Callable:
    def __init__(self, anycallable):
        self.__call__ = anycallable 

class CBD_PlotterController:
  instance = None

  def __init__(self, model=None):
    if CBD_PlotterController.instance != None:
      raise RunTimeError("Only one instance of this controller is allowed")    
    self.__plotters = []
    self.model      = model #must be externally setted (unclean, due to plotter)

    
  def getInstance():
    if CBD_PlotterController.instance == None:    
      CBD_PlotterController.instance = CBD_PlotterController()
    return CBD_PlotterController.instance             
  getInstance=Callable(getInstance)    

  def addObserver(self, plotter):
    self.__plotters.append(plotter)

  def removeObserver(self, plotter):
    if plotter in self.__plotters:
      self.__plotters.remove(plotter)
    else:
      raise ValueError("Specified plotter is invalid")

  def notify(self):
    for plotter in self.__plotters:
      plotter.update()
      
  def __convertList(self, atomList):
    newList=[]
    for elem in atomList:
      newList.append(elem.getValue())
    return newList
    
  def __getSignals(self):
    signalTypes = [BLOCK_SIGNAL, TIME_SIGNAL, TERMINATE_SIGNAL]
    signals = []
    for signalType in signalTypes:
      if self.model.listNodes.has_key(signalType):
        signals+= self.model.listNodes[signalType]
      else:
        raise ValueError("Incorrect Signal Type")
    return signals                
    
  def getTitles(self):
    signals = self.__getSignals()
    titles  = []
    for signal in signals:
      if signal.toPlot.getValue()[1]:
        name  = signal.signal_name.getValue()     
        titles.append(name)
                 
    titles.sort() 
    return titles
                  
  def getCoords(self, xTitle, yTitle): 
    signals = self.__getSignals()
    state = [None, None]
    for signal in signals:
      name  = signal.signal_name.getValue()    
      if xTitle==name:
        state[0]=self.__convertList(signal.signal.getValue())
      if yTitle==name:
        state[1]=self.__convertList(signal.signal.getValue())
           
    return state
    
  def __str__(self):
    return "Causal Block Diagram Plotting Controller"
      
