### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
##                    308-522 -- Modelling and Simulation
##                                  Fall 2004
##                             --- ASSIGNMENT 1 ---
##
## CBD_genLaTeX.py
##
## last modified: 09/27/04
### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###

from CBD_Constants import *
import tkFileDialog  # For browser dialog
from CBD_Simulator import TS_Simulator

# List of blocks in ascending order of precedence
PRECEDENCE_LIST = [ADDER_BLOCK,
                  PRODUCT_BLOCK,
                  NEGATOR_BLOCK,
                  INVERTER_BLOCK,
                  CONSTANT_BLOCK,
                  GENERIC_BLOCK,
                  FILE_IO_BLOCK]

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def genLaTeX(parentApp, model):
    sim = TS_Simulator(model)
    sim.checkLegality()

    #initilizations
    timeDelayBlocks = []  # list of delay blocks(derivative, integrator, and delay)
    constantBlocks = []   # list of blocks of type constants with non-empty names
    testBlocks = []       # list of all named test blocks
    otherBlocks = []      # list of all other named blocks but Generic and FileI/O
    output = ""           # The output string

    # List of all blocks:
    blocks = model.listNodes[BLOCK]

    for b in blocks:
        blockType = sim.getBlockType(b)
        blockName = sim.getBlockName(b)

        if ( (blockType == DERIVATIVE_BLOCK) or
             (blockType == INTEGRATOR_BLOCK) or
             (blockType == DELAY_BLOCK) ):
            if (blockName == ""):
               print "Error: Derivative, Integrator and Delay blocks must be named!"
               return None
            timeDelayBlocks.append(b)
        elif (blockType == TEST_BLOCK):
            if (blockName != ""):
               testBlocks.append(b)
            else:
               print "Error: Test blocks must be named!"
               return None
        elif (blockType == CONSTANT_BLOCK):
            if (blockName != ""):
               constantBlocks.append(b)
        elif (blockType != GENERIC_BLOCK and
              blockType != FILE_IO_BLOCK):
             if (blockName != ""):
                otherBlocks.append(b)

    # write the LaTeX decoration
    output += "\documentclass[12pt] {article}\n"
    output += "\usepackage{pslatex}\n"
    output += "\\begin{document}\n"
    output += "\[\n"
    output += model.model_name.getValue() + " = \n"
    output += "  \left\{\n"
    output += "    \\begin{array}{ll}\n"

    # Start writing the equation:
    # 1- Build a list of the blocks that are unnamed IC for time delay blocks
    # 2- Write the time delay block equations
    icEqs = []
    for b in timeDelayBlocks:
        # Keep a list of all blocks for which to write an equation and their
        # IC block
        ic = b.block_IC_port[0].in_connections_[0]
        if (ic.getClass() != SIMCONTROL and sim.getBlockName(ic) == ""):
           icEqs.append([b, ic])

        tmp = printBlock(sim, b)
        if(tmp == None):
            return
        output += ("      " + tmp + "\\\\\n")

    if (len(timeDelayBlocks) > 0):
       output += "      \\\\\n"

    # 3- Write the test blocks equations
    for b in testBlocks:
        # We know that the name is not empty. Because otherwise the block
        # wouldn't have been in this list.
        equation = "      " + sim.getBlockName(b) + " = "
        tmp = printBlock(sim, b)
        if (tmp == None):
           return
        else:
           equation += tmp
        output += equation

    if (len(testBlocks) > 0):
       output += "      \\\\\n      \\\\\n"

    # 4- Write the other blocks equations, except constants and IC blocks.
    for b in otherBlocks:
        # We know that the name is not empty. Because otherwise the block
        # wouldn't have been in this list.
        equation = "      " + sim.getBlockName(b) + " = "
        tmp = printBlock(sim, b)
        if (tmp == None):
           return
        else:
           equation += tmp
        equation += "\\\\\n"
        output += equation

    # 5- Write the equations for unnamed IC blocks
    for block in icEqs:
        output += ("      " +
              sim.getBlockName(block[0]) +
              "{_0} = " +
              printBlock(sim, block[1]) +
              "\\\\\n")

    # 6- Write the constant value equations
    for b in constantBlocks:
        if (sim.getBlockName(b) != ""):
            output += ("      " +
                    sim.getBlockName(b) +
                    " = " +
                    str(b.block_out_value.getValue()) +
                    "\\\\\n")

    # More LaTeX decorations
    output += "    \end{array}\n"
    output += "  \\right.\n"
    output += "\]\n"
    output += "\end{document}\n"

    print output
    writeToFile(output)

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# This function prints the equation of a block by calling its appropriate print
# function.
def printBlock(sim, block):
    blockType = sim.getBlockType(block)

    if (blockType == NEGATOR_BLOCK):
       output = printNegator(sim, block)
    elif (blockType == INVERTER_BLOCK):
       output = printInverter(sim, block)
    elif (blockType == ADDER_BLOCK):
       output = printAdder(sim, block)
    elif (blockType == PRODUCT_BLOCK):
       output = printProduct(sim, block)
    elif (blockType == DELAY_BLOCK):
       output = printDelay(sim, block)
    elif (blockType == INTEGRATOR_BLOCK):
       output = printIntegrator(sim, block)
    elif (blockType == DERIVATIVE_BLOCK):
       output = printDerivative(sim, block)
    elif (blockType == CONSTANT_BLOCK):
       output = printConstant(sim, block)
    elif (blockType == TEST_BLOCK):
       output = printTest(sim, block)
    elif (blockType == GENERIC_BLOCK):
       output = printGeneric(sim, block)
    elif (blockType == FILE_IO_BLOCK):
       print "Error: Cannot evaluate FileIO blocks!"
       return
    else:
       print ("Error: Unknown block type!")
       return

    if (output == None):
          return

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For a generic block, just print its name and determine the equation of all
# its input blocks. Since, generic block has the lowest precedence rule among
# all blocks, we're going to add the parantheses without checking the precedence
# rules.
def printGeneric(sim, block):
    output = ""
    if (sim.getBlockName(block) != ""):
        output += sim.getBlockName(block) + "("
        for link in block.in_connections_:
            for b in link.in_connections_:
	    	if (b.getClass() == SIMCONTROL):
         	     output += printSimControl(b, block.in_connections_[0])
		else:
                     output += printBlock(sim, b)
        output += ")"

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For constant blocks, just print their value.
def printConstant(sim, block):
    if (sim.getBlockName(block) != ""):
        return sim.getBlockName(block)
    else:
        return str(block.block_out_value.getValue())

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For negator blocks, print a minus sign ('-') and then evaluate the equation
# of its input node.
def printNegator(sim, block):
    output = "-"
    inputBlock = block.in_connections_[0].in_connections_[0]

    if (inputBlock.getClass() == SIMCONTROL):
         output += printSimControl(inputBlock, block.in_connections_[0])
    elif (sim.getBlockName(inputBlock) != ""):
         output += sim.getBlockName(inputBlock)
    else:
         output += "("
         tmp = printBlock(sim, inputBlock)
         if(tmp == None):
            return
         output += tmp + ")"

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For invertors, print a "1 / " and then evaluate the equation of its input node.
def printInverter(sim, block):
    output = "\\frac{1}"

    inputBlock = block.in_connections_[0].in_connections_[0]
    inputIsNamed = False
    if (inputBlock.getClass() == SIMCONTROL):
         tmp = printSimControl(inputBlock, block.in_connections_[0])
	 inputIsNamed = True
    elif (sim.getBlockName(inputBlock) != ""):
        tmp = sim.getBlockName(inputBlock)
        inputIsNamed = True
    else:
        tmp = printBlock(sim, inputBlock)

    if(tmp == None):
        return

    # Check if the operator of the input block to this inverter has a lower
    # precedence. If it does, then put parantheses around the input operation.
    if (not inputIsNamed and
        PRECEDENCE_LIST.index(sim.getBlockType(inputBlock)) < PRECEDENCE_LIST.index(INVERTER_BLOCK)):
          output += "{(" + tmp + ")}"
    else:
          output += "{" + tmp + "}"

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For adders, evaluate all of its input nodes and add a '+' between them.
def printAdder(sim, block):
    output = ""    # initialize result to empty string
    index = 0
    for link in block.in_connections_:
        for node in link.in_connections_:
	    inputIsNamed = False
	    if (node.getClass() == SIMCONTROL):
           	tmp = printSimControl(node, block.in_connections_[index])
		inputIsNamed = True
            elif (node.getClass() == BLOCK):
        	if (sim.getBlockType(node) != GENERIC_BLOCK and
                    sim.getBlockName(node) != ""):
           		tmp = sim.getBlockName(node)
           		inputIsNamed = True
        	else:
             	     tmp = printBlock(sim, node)

             	if (tmp == None):
                   return

            # If the input block is not named and its operation has a precedence
            # lower than this block operation, we need to put parantheses around it.
            if (not inputIsNamed and
                PRECEDENCE_LIST.index(sim.getBlockType(node)) < PRECEDENCE_LIST.index(ADDER_BLOCK)):
             	  output += "(" + tmp + ")"
            else:
             	  output += tmp

	    # If this is not the last input node print the plus sign
            if (index < len(block.in_connections_) - 1):
               	output += " + "
	    index += 1

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For product blocks, evaluate all of its input nodes and add a '*' between them.
def printProduct(sim, block):
    output = ""    # initialize result to empty string
    index = 0
    for link in block.in_connections_:
        for node in link.in_connections_:
	    inputIsNamed = False
	    if (node.getClass() == SIMCONTROL):
           	tmp = printSimControl(node, block.in_connections_[index])
		inputIsNamed = True
            elif (node.getClass() == BLOCK):
        	if (sim.getBlockType(node) != GENERIC_BLOCK and 
                    sim.getBlockName(node) != ""):
           		tmp = sim.getBlockName(node)
           		inputIsNamed = True
        	else:
             	     tmp = printBlock(sim, node)

             	if (tmp == None):
                   return

            # If the input block is not named and its operation has a precedence
            # lower than this block operation, we need to put parantheses around it.
            if (not inputIsNamed and
                PRECEDENCE_LIST.index(sim.getBlockType(node)) < PRECEDENCE_LIST.index(ADDER_BLOCK)):
             	  output += "(" + tmp + ")"
            else:
             	  output += tmp

	    # If this is not the last input node print the plus sign
            if (index < len(block.in_connections_) - 1):
               	output += " \cdot "
	    index += 1

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def printTest(sim, block):
    inputBlock = sim.getInputSignals(block)[0].in_connections_[0]
    trueBlock = block.block_TRUE_in_port[0].in_connections_[0]
    falseBlock = block.block_FALSE_in_port[0].in_connections_[0]

    if (inputBlock.getClass() == SIMCONTROL):
        inputBlockName = printSimControl(inputBlock, block.block_in_port[0])
    else:
    	inputBlockName = sim.getBlockName(inputBlock)
    print("input="+inputBlockName)
    if (trueBlock.getClass() == SIMCONTROL):
        trueBlockName = printSimControl(trueBlock, block.block_TRUE_in_port[0])
    else:
    	trueBlockName = sim.getBlockName(trueBlock)
    if (falseBlock.getClass() == SIMCONTROL):
        falseBlockName = printSimControl(falseBlock, block.block_FALSE_in_port[0])
    else:
    	falseBlockName = sim.getBlockName(falseBlock)

    output = "\n" + "      \left\{\n        \\begin{array}{ll}\n          "

    if (trueBlockName != ""):
       output += trueBlockName
    else:
       tmp = printBlock(sim, trueBlock)
       if (tmp == None):
          return
       output += tmp

    output += ", & if "

    if (inputBlockName != ""):
       output += inputBlockName
    else:
       tmp = printBlock(sim, inputBlock)
       if (tmp == None):
          return
       output += tmp

    output += " > 0\\\\\n          "

    if (falseBlockName != ""):
       output += falseBlockName
    else:
       tmp = printBlock(sim, falseBlock)
       if (tmp == None):
          return
       output += tmp

    output += ", & if "

    if (inputBlockName != ""):
       output += inputBlockName
    else:
       tmp = printBlock(sim, inputBlock)
       if (tmp == None):
          return
       output += tmp

    output += " <= 0\\\\\n        \end{array}\n      \\right.\n"

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def printIntegrator(sim, block):
    name = sim.getBlockName(block)
    output = "\dot{" + name + "} = "

    inOperator = sim.getInputSignals(block)[0].in_connections_[0]
    if (inOperator.getClass() == SIMCONTROL):
        output += printSimControl(inOperator, block.in_connections_[1])
    elif (sim.getBlockType(inOperator) == INTEGRATOR_BLOCK or
          sim.getBlockType(inOperator) == DERIVATIVE_BLOCK or
          sim.getBlockType(inOperator) == DELAY_BLOCK):
        output += sim.getBlockName(inOperator)
    else:
        tmp = printBlock(sim, inOperator)
        if(tmp == None):
            return
        output += tmp

    output += ", & " + name + "(t_0) = "
    IC_name = sim.getBlockName(block.block_IC_port[0].in_connections_[0])
    if (IC_name != ""):
        output += IC_name
    else:
       output += name
       output += "{_0}"

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def printDerivative(sim, block):
    name = sim.getBlockName(block)
    output = name + "^{\prime}" + " = "

    inOperator = sim.getInputSignals(block)[0].in_connections_[0]
    if (inOperator.getClass() == SIMCONTROL):
        output += printSimControl(inOperator, block.in_connections_[1])
    elif (sim.getBlockType(inOperator) == INTEGRATOR_BLOCK or
          sim.getBlockType(inOperator) == DERIVATIVE_BLOCK or
          sim.getBlockType(inOperator) == DELAY_BLOCK):
        output += sim.getBlockName(inOperator)
    else:
        tmp = printBlock(sim, inOperator)
        if(tmp == None):
            return
        output += tmp

    output += ", & " + name + "(t_0) = "
    IC_name = sim.getBlockName(block.block_IC_port[0].in_connections_[0])
    if (IC_name != ""):
        output += IC_name
    else:
       output += name
       output += "{_0}"

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def printDelay(sim, block):
    name = sim.getBlockName(block)
    IC = block.block_IC_port[0].in_connections_[0]
    inputBlock = sim.getInputSignals(block)[0].in_connections_[0]
    return ( name + "_t = " + printBlock(sim, inputBlock) + "_{t-1}," +
             " \\forall t > 0, & " + name + "_0 = " +  printBlock(sim, IC))

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# This function determines whether the given signal is a simControl_discreteTime
# simControl_dcontinuousTime, or simControl_timeIncrement.
def printSimControl(simcontrol, signal):
    if (signal in simcontrol.simControl_discreteTime_port):
       return "simControlDiscreteTime"
    elif (signal in simcontrol.simControl_continuousTime_port):
       return "simControlContinuousTime"
    elif (signal in simcontrol.simControl_timeIncrement_port):
       return "simControlTimeIncrement"

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def writeToFile(string):
    filename = tkFileDialog.asksaveasfilename()
    if not filename:
       return

    # Make sure the name is valid:
    assert (filename[-4:] == ".tex"), "invalid name"

    #  writ to the file
    fd = open(filename, 'w')
    fd.write(string)
    fd.close()
    
    print ("The equations in LaTeX format have been written to " + filename)
