from CBD_Constants import *
from CBD_DepGraph import DepGraph

import os
import os.path
import string
import tkFileDialog
import distutils.file_util

BASEDIR = os.path.split( __file__ )[ 0 ]   

def genMDC( model ):
  checkLegality(model)

  output = ""

  blocks = model.listNodes[BLOCK]
    
  output += "# Casual Block Class" + chr( 10 )
  output += chr( 10 )
  
  output += "from math import *"+ chr( 10 )
  output += chr( 10 )
  output += "from BaseBlock import *"+ chr( 10 )
  output += chr( 10 )
  
  importList = []
  for block in blocks:
    blockType = getBlockType( block )
    blockOperatorName = block.block_operator.getValue()
    if ( blockType == CHILD_BLOCK ):
      try:
        importList.index( blockOperatorName )
      except:
        importList.append( blockOperatorName )
  
  for importClass in importList:
    output += "from " + importClass + " import *" + chr( 10 )
  output += chr( 10 )      

  
  output += "class " + model.model_name.getValue() + ":" + chr( 10 )
  output += chr( 10 )
  output += "  def __init__(self, initParams={}):" + chr( 10 )
  output += "    self.__initParams = initParams" + chr( 10 )
  output += "    self.__blocks = []" + chr( 10 )
  output += chr( 10 )
  
  output += "    # build blocks" + chr( 10 )
  for block in blocks:
    blockType = getBlockType( block )
    blockName = getBlockName( block )    
    
    if ( blockType == GENERIC_BLOCK ):
      output += "    self." + blockName + " = GenericBlock(\"" + blockName + "\", \"" + blockType + "\", \"" + \
                block.block_operator.getValue() + "\")"                
    elif ( blockType == CONSTANT_BLOCK ):
      output += "    self." + blockName + " = ConstantBlock(\"" + blockName + "\", \"" + blockType + "\", " + \
                str( block.block_out_value.getValue() ) + ")"      
    elif ( blockType == CHILD_BLOCK ): 
      output += "    if (self.__initParams.has_key(\"" + blockName + "\") == 1):" + chr( 10 )
      output += "      self." + blockName + " = ChildBlock(\"" + blockName + "\", \"" + blockType + "\", " + \
                block.block_operator.getValue() + "(self.__initParams[\"" + blockName + "\"]))" + chr( 10 )
      output += "    else:" + chr( 10 )      
      output += "      self." + blockName + " = ChildBlock(\"" + blockName + "\", \"" + blockType + "\", " + \
                block.block_operator.getValue() + "())"
    elif ( blockType == TEST_BLOCK ): 
      output += "    self." + blockName + " = TestBlock(\"" + blockName + "\", \"" + blockType + "\")"
    elif ( blockType == DELAY_BLOCK or 
          blockType == DERIVATIVE_BLOCK or 
          blockType == INTEGRATOR_BLOCK ): 
      output += "    self." + blockName + " = TimeDelayBlock(\"" + blockName + "\", \"" + blockType + "\", " + \
                str( model.simul_delta_t.getValue() ) + ")"   
    elif ( blockType == NEGATOR_BLOCK or 
          blockType == INVERTER_BLOCK or 
          blockType == ADDER_BLOCK or 
          blockType == PRODUCT_BLOCK or 
          blockType == TIME_BLOCK or 
          blockType == INPORT_BLOCK or 
          blockType == OUTPORT_BLOCK ): 
      output += "    self." + blockName + " = BaseBlock(\"" + blockName + "\", \"" + blockType + "\")"
    elif ( blockType == FILE_IO_BLOCK ):
      raise ValueError( "Cannot treat FileIO blocks" )   
    output += chr( 10 )
    
    output += "    self.__blocks.append(self."+blockName+")"
    output += chr( 10 )

  output += chr( 10 )
  output += "    # build connections" + chr( 10 )
  for block in blocks:
    blockType = getBlockType( block )
    blockName = getBlockName( block )
    for in_connection in block.in_connections_:
      in_block_name = getBlockName( in_connection.in_connections_[0] )
      connection_name = in_connection.signal_name.getValue()
      inport_name = in_connection.inport_name.getValue()
      outport_name = in_connection.outport_name.getValue()
      output += "    self." + blockName + ".linksIN.append((self." + in_block_name + \
                ", \"" + connection_name + "\", \"" + inport_name + "\", \"" + outport_name + "\"))" + chr( 10 )
    if ( blockType == TEST_BLOCK ):
      trueBlockName = getBlockName( block.block_TRUE_in_port[0].in_connections_[0] )
      falseBlockName = getBlockName( block.block_FALSE_in_port[0].in_connections_[0] )
      truePortName = block.block_TRUE_in_port[0].inport_name.getValue()
      falsePortName = block.block_FALSE_in_port[0].inport_name.getValue()
      output += "    self." + blockName + ".TRUE_in = (self." + trueBlockName + \
                ", \"" + block.block_TRUE_in_port[0].signal_name.getValue() + \
                "\", \"" + truePortName + "\", \"\")" + chr( 10 )
      output += "    self." + blockName + ".FALSE_in = (self." + falseBlockName + \
                ", \"" + block.block_FALSE_in_port[0].signal_name.getValue() + \
                "\", \"" + falsePortName + "\", \"\")" + chr( 10 )
    elif ( blockType == DELAY_BLOCK or 
          blockType == DERIVATIVE_BLOCK or 
          blockType == INTEGRATOR_BLOCK ):
      icBlockName = getBlockName( block.block_IC_port[0].in_connections_[0] )
      icPortName = block.block_IC_port[0].inport_name.getValue()
      output += "    self." + blockName + ".IC = (self." + icBlockName + \
                ", \"" + block.block_IC_port[0].signal_name.getValue() + \
                "\", \"" + icPortName + "\", \"\")" + chr( 10 )

  output += chr( 10 )
  output += "    # parameters initialization" + chr( 10 )
  output += "    for block in self.__blocks:" + chr( 10 )
  output += "      if isinstance(block, ConstantBlock):" + chr( 10 )
  output += "        if (self.__initParams.has_key(block.getBlockName()) == 1):" + chr( 10 )
  output += "          block.init_value = self.__initParams[block.getBlockName()]" + chr( 10 )
  output += "        " + chr( 10 )

                    
  output += chr( 10 )
  output += "  # to access all blocks outside the model" + chr( 10 )
  output += "  def getBlocks(self):" + chr( 10 )
  output += "    return self.__blocks" + chr( 10 )  

#  print output
  writeToFile( model.model_name.getValue(), output )
  
def getBlockType(block):
  if __getType(block)==BLOCK:
    index = block.block_type.getValue()[1]
    return block.block_type.getValue()[0][index]
  else:
    raise ValueError("Expected a Block node")
  
def getBlockName(block):
  if __getType(block)==BLOCK:
    return block.block_name.getValue()
  else:
    raise ValueError("Expected a Block node")

def __getType(ASGNode):
  return ASGNode.getTypeName()  

# Get the signals of a block that are neither IC, FALSE, or TRUE input ports
def getInputSignals(block):
  answer = []
  for signal in block.in_connections_:
    if (signal not in block.block_IC_port and signal not in block.block_TRUE_in_port and signal not in block.block_FALSE_in_port):
      answer.append(signal)
  return answer 

  # Traverses the CBD and checks that all blocks are legally connected
def checkLegality(model):
  blocks = model.listNodes[BLOCK]
  for block in blocks:
    blockType = getBlockType(block)
    blockName = getBlockName(block)
    if (blockName == ""):
        blockName = blockType
    # for negator, inverter, delay, derivative, integrator, and test blocks, one and only one input signal is allowed
    if (blockType == NEGATOR_BLOCK or blockType == INVERTER_BLOCK or blockType == DELAY_BLOCK or blockType == DERIVATIVE_BLOCK or blockType == INTEGRATOR_BLOCK or blockType == TEST_BLOCK):
      assert len(getInputSignals(block)) == 1, blockName + " has invalid number of inputs"
    # for delay, derivative and integrator blocks, one and only one IC input is allowed
    if (blockType == DELAY_BLOCK or blockType == DERIVATIVE_BLOCK or blockType == INTEGRATOR_BLOCK):
      assert len(block.block_IC_port) == 1, blockName + " has invalid number of initial conditions"
    # for test blocks, one and only one TRUE input is allowed, and one and only one FALSE input is allowed
    if (blockType == TEST_BLOCK):
      assert len(block.block_TRUE_in_port) == 1, blockName + " has invalid number of inputs (TRUE input)"
      assert len(block.block_FALSE_in_port) == 1, blockName + " has invalid number of inputs (FALSE input)"
  
def writeToFile( modelname, string ):  
  IO = getIO()
  IO.askPath()
  IO.copySimulationEnvironment() 
   
  file = getIO().fopen( modelname )  
  file.write( string )
  file.close()
    
  print ( "The Casual Block Class has been written to " + modelname+".py" )

def getIO():
  if( IOstuff.INSTANCE == None ):
    IOstuff.INSTANCE = IOstuff()
  return IOstuff.INSTANCE

# Extended from part of Ernesto Posse's solution for DEVS model code generation:
# http://msdl.cs.mcgill.ca/MSDL/people/eposse/projects.dtml  
class IOstuff:
  """
  Handles file paths, singleton class
  """
  INSTANCE = None
  
  def __init__( self ):
    self.basePath = os.path.join( BASEDIR, 'Generated_code_dump' )
  
  def __cleanPath( self, path ):
    self.path = os.path.normpath( path )
    return path
  
  def askPath( self ):
    """ Get the user to supply the directory """
    try:  
      pathFile = tkFileDialog.asksaveasfilename( 
                  title="Where do you want your Causal Block Class code? (EMPTY DIR)", 
                  filetypes=[( "Select an empty directory", "*" )], 
                  initialfile='Ignored'
                  , initialdir = __file__ )   
    except:
      pathFile = tkFileDialog.askdirectory( 
                title="Where do you want your Causal Block Class code? (EMPTY DIR)", 
                initialdir=__file__ )                          
      if( pathFile ): return self.__cleanPath( os.path.split( pathFile ) )
    if( pathFile ): return self.__cleanPath( os.path.split( pathFile )[0] )
    return self.__cleanPath( os.path.split( self.basePath )[0] )
    
  def setPath( self, path, name ):
    self.path = os.path.join( path, name )
    return self.path
  
  def getPath( self ):
    return self.path
  
  def fopen( self, name ):
    # Make the neccessary directories
    if( not os.path.exists( self.path ) ):
      os.makedirs( self.path )
    return open( os.path.join( self.path, name + ".py" ), "w" )
  
  def copySimulationEnvironment( self ):
    # Make the neccessary directories
    if( not os.path.exists( self.path ) ):
      os.makedirs( self.path )
    d = os.path.join( BASEDIR, 'ST_CBD' )
    for file in os.listdir( d ):
      distutils.file_util.copy_file( os.path.join( d, file ), 
                                    os.path.join( self.path, file ) )          