### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
##                    308-522 -- Modelling and Simulation
##                                  Fall 2005
##                             --- ASSIGNMENT 1 ---
##
## GenLaTeX.py : Generate a latex file of a CBD model (for flat models only now).
## 
## For this assignment, you need to implement the following missing functions:
##      printNegator(block)
##      printAdder(block)
##      printProduct(block)
##      printDelay(block)
##      printIntegrator(block)
##      printDerivative(block)
##      printTest(block)
##   
## 
## last modified: 09/29/05
### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###

import tkFileDialog  # For browser dialog

from BaseBlock import *

GENERIC_BLOCK    = "Generic"
NEGATOR_BLOCK    = "Negator"
INVERTER_BLOCK   = "Inverter"
ADDER_BLOCK      = "Adder"
PRODUCT_BLOCK    = "Product"
DELAY_BLOCK      = "Delay"
INTEGRATOR_BLOCK = "Integrator"
DERIVATIVE_BLOCK = "Derivative"
CONSTANT_BLOCK   = "Constant"
TEST_BLOCK       = "Test"

INPORT_BLOCK     = "Inport"
OUTPORT_BLOCK    = "Outport"
CHILD_BLOCK      = "Child"

# List of blocks in ascending order of precedence
PRECEDENCE_LIST = [ADDER_BLOCK,
                  PRODUCT_BLOCK,
                  NEGATOR_BLOCK,
                  INVERTER_BLOCK,
                  CONSTANT_BLOCK,
                  GENERIC_BLOCK]

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def genLaTeX(model, model_name):

    #initilizations
    timeDelayBlocks = []  # list of delay blocks(derivative, integrator, and delay)
    constantBlocks = []   # list of blocks of type constants with non-empty names
    testBlocks = []       # list of all named test blocks
    otherBlocks = []      # list of all other named blocks but Generic and FileI/O
    output = ""           # The output string

    # List of all blocks:
    blocks = model.getBlocks()

    for b in blocks:
        blockType = b.getBlockType()
        blockName = b.getBlockName()

        if ( (blockType == DERIVATIVE_BLOCK) or
             (blockType == INTEGRATOR_BLOCK) or
             (blockType == DELAY_BLOCK) ):
            if (blockName == ""):
               print "Error: Derivative, Integrator and Delay blocks must be named!"
               return None
            timeDelayBlocks.append(b)
        elif (blockType == TEST_BLOCK):
            if (blockName != ""):
               testBlocks.append(b)
            else:
               print "Error: Test blocks must be named!"
               return None
        elif (blockType == CONSTANT_BLOCK):
            if (not blockName.startswith("_")):
               constantBlocks.append(b)
        elif (blockType != GENERIC_BLOCK):
             if (not blockName.startswith("_")):
                otherBlocks.append(b)
        elif (blockType == INPORT_BLOCK or blockType == OUTPORT_BLOCK or blockType == CHILD_BLOCK):
             print "Error: Child, Inport and Outport blocks for hierarchical modelling are not supported!"
             return None

    # write the LaTeX decoration
    output += "\documentclass[12pt] {article}\n"
    output += "\usepackage{pslatex}\n"
    output += "\\begin{document}\n"
    output += "\[\n"
    output += model_name + " = \n"
    output += "  \left\{\n"
    output += "    \\begin{array}{ll}\n"

    # Start writing the equation:
    # 1- Build a list of the blocks that are unnamed IC for time delay blocks
    # 2- Write the time delay block equations
    icEqs = []
    for b in timeDelayBlocks:
        # Keep a list of all blocks for which to write an equation and their
        # IC block
        ic = b.IC[0]
        if (ic.getBlockName().startswith("_")):
           icEqs.append([b, ic])

        tmp = printBlock(b)
        if(tmp == None):
            return
        output += ("      " + tmp + "\\\\\n")

    if (len(timeDelayBlocks) > 0):
       output += "      \\\\\n"

    # 3- Write the test blocks equations
    for b in testBlocks:
        # We know that the name is not empty. Because otherwise the block
        # wouldn't have been in this list.
        equation = "      " + b.getBlockName() + " = "
        tmp = printBlock(b)
        if (tmp == None):
           return
        else:
           equation += tmp
        output += equation

    if (len(testBlocks) > 0):
       output += "      \\\\\n      \\\\\n"

    # 4- Write the other blocks equations, except constants and IC blocks.
    for b in otherBlocks:
        # We know that the name is not empty. Because otherwise the block
        # wouldn't have been in this list.
        equation = "      " + b.getBlockName() + " = "
        tmp = printBlock(b)
        if (tmp == None):
           return
        else:
           equation += tmp
        equation += "\\\\\n"
        output += equation

    # 5- Write the equations for unnamed IC blocks
    for blocklist in icEqs:
        output += ("      " +
              blocklist[0].getBlockName() +
              ",{_0} = " +
              printBlock(blocklist[1]) +
              "\\\\\n")

    # 6- Write the constant value equations
    for b in constantBlocks:
        if (not b.getBlockName().startswith("_")):
            output += ("      " +
                    b.getBlockName() +
                    " = " +
                    str(b.init_value) +
                    "\\\\\n")

    # More LaTeX decorations
    output += "    \end{array}\n"
    output += "  \\right.\n"
    output += "\]\n"
    output += "\end{document}\n"

    print output
    writeToFile(output)

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# This function prints the equation of a block by calling its appropriate print
# function.
def printBlock(block):
    blockType = block.getBlockType()

    if (blockType == NEGATOR_BLOCK):
       output = printNegator(block)
    elif (blockType == INVERTER_BLOCK):
       output = printInverter(block)
    elif (blockType == ADDER_BLOCK):
       output = printAdder(block)
    elif (blockType == PRODUCT_BLOCK):
       output = printProduct(block)
    elif (blockType == DELAY_BLOCK):
       output = printDelay(block)
    elif (blockType == INTEGRATOR_BLOCK):
       output = printIntegrator(block)
    elif (blockType == DERIVATIVE_BLOCK):
       output = printDerivative(block)
    elif (blockType == CONSTANT_BLOCK):
       output = printConstant(block)
    elif (blockType == TEST_BLOCK):
       output = printTest(block)
    elif (blockType == GENERIC_BLOCK):
       output = printGeneric(block)
    else:
       print ("Error: Unknown block type!")
       return

    if (output == None):
          return

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For negator blocks, print a minus sign ('-') and then evaluate the equation
# of its input node.
def printNegator(block):
    output = "-"

    (inputBlock, s, inputn, outputn) = block.linksIN[0]
    inputIsNamed = False
    if (not inputBlock.getBlockName().startswith("_")):
        tmp = inputBlock.getBlockName()
        inputIsNamed = True
    else:
        tmp = printBlock(inputBlock)

    if(tmp == None):
        return

    # Add parentheses if needed
    if (not inputIsNamed and
        PRECEDENCE_LIST.index(inputBlock.getBlockType()) < PRECEDENCE_LIST.index(NEGATOR_BLOCK)):
          output += "(" + tmp + ")"
    else:
          output += tmp

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For adders, evaluate all of its input nodes and add a '+' between them.
def printAdder(block):
    output = ""
    numOfSign = len(block.linksIN) - 1
    
    for (b, s, inputn, outputn) in block.linksIN:
      blockName = b.getBlockName()
      if(not blockName.startswith("_")):
        tmp = b.getBlockName()
      else:
        tmp = printBlock(b)
      
      if (tmp == None):
        return
      else:
        output += tmp
        
      if (numOfSign > 0):
        output += " + "
        numOfSign -= 1
        
    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For product blocks, evaluate all of its input nodes and add a '*' between them.
def printProduct(block):
    output = ""
    numOfSign = len(block.linksIN) - 1
    
    for (b, s, inputn, outputn) in block.linksIN:
      blockName = b.getBlockName()
      inputIsNamed = False
      if(not blockName.startswith("_")):
        tmp = b.getBlockName()
        inputIsNamed = True
      else:
        tmp = printBlock(b)
      
      # Add parentheses if needed
      if (tmp == None):
        return
      elif (not inputIsNamed and b.getBlockType() == ADDER_BLOCK):
        output += "(" + tmp + ")"
      else:
        output += tmp
        
      if (numOfSign > 0):
        output += " \cdot "
        numOfSign -= 1
        
    return output
    
### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For test blocks, evaluate input node, true_in node and false_in node and
# print "= true_in, if:input > 0,    = false_in, otherwise" as output
def printTest(block):
    inputBlock = getInputBlocks(block)[0]
    trueBlock = block.TRUE_in[0]
    falseBlock = block.FALSE_in[0]

    inputBlockName = inputBlock.getBlockName()
    trueBlockName = trueBlock.getBlockName()
    falseBlockName = falseBlock.getBlockName()

    output = ""
    if (not trueBlockName.startswith("_")):
      tmp = trueBlockName 
    else:
      tmp = printBlock(trueBlock)
      
    if (tmp == None):
      return
    else:
      output += tmp + ", if:" 
      
    if (not inputBlockName.startswith("_")):
      tmp = inputBlockName 
    else:
      tmp = printBlock(inputBlock) 

    if (tmp == None):
      return
    else:
      output += tmp + " > 0, & = "

    if (not falseBlockName.startswith("_")):
      tmp = falseBlockName
    else:
      tmp = printBlock(falseBlock)  
      
    if (tmp == None):
      return
    else:
      output += tmp + ", otherwise"
      
    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def printIntegrator(block):
    output = "\dot{" + block.getBlockName() + "} = "
    
    inputBlock = getInputBlocks(block)[0]
    icBlock = block.IC[0]
    
    if (not inputBlock.getBlockName().startswith("_")):
      output += inputBlock.getBlockName()
    else:
      output += printBlock(inputBlock)
      
    output += ", & "
    output += block.getBlockName() + "(t_0) = "
    
    if (not icBlock.getBlockName().startswith("_")):
      output += icBlock.getBlockName()
    else:
      output += block.getBlockName() + ",{_0}"
    
    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def printDerivative(block):
    output = "$" + block.getBlockName() + "'$ = "
    
    inputBlock = getInputBlocks(block)[0]
    icBlock = block.IC[0]
    
    if (not inputBlock.getBlockName().startswith("_")):
      output += inputBlock.getBlockName()
    else:
      output += printBlock(inputBlock)
      
    output += ", & "
    output += block.getBlockName() + "(t_0) = "
    
    if (not icBlock.getBlockName().startswith("_")):
      output += icBlock.getBlockName()
    else:
      output += block.getBlockName() + ",{_0}"
    
    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def printDelay(block):
    output = block.getBlockName() + "(t) = "
    
    inputBlock = getInputBlocks(block)[0]
    icBlock = block.IC[0]
    
    if (not inputBlock.getBlockName().startswith("_")):
      output += inputBlock.getBlockName() + "(t - t_d)"
    else:
      output += "(" + printBlock(inputBlock) + ")(t - t_d)"
      
    output += ", & "
    output += block.getBlockName() + "(t_0) = "
    
    if (not icBlock.getBlockName().startswith("_")):
      output += icBlock.getBlockName()
    else:
      output += block.getBlockName() + ",{_0}"
    
    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For a generic block, just print its name and determine the equation of all
# its input blocks. Since, generic block has the lowest precedence rule among
# all blocks, we're going to add the parantheses without checking the precedence
# rules.
def printGeneric(block):
    output = ""
    if (not block.getBlockName().startswith("_")):
        output += block.block_operator + "("
        for (b, s, inputn, outputn) in block.linksIN:
                output += printBlock(b)
        output += ")"

    return output

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For constant blocks, just print their value.
def printConstant(block):
    if (not block.getBlockName().startswith("_")):
        return block.getBlockName()
    else:
        return str(block.init_value)

### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
# For invertors, print a "1 / " and then evaluate the equation of its input node.
def printInverter(block):
    output = "\\frac{1}"

    (inputBlock, inputs, inputn, outputn) = block.linksIN[0]
    inputIsNamed = False
    if (not inputBlock.getBlockName().startswith("_")):
        tmp = inputBlock.getBlockName()
        inputIsNamed = True
    else:
        tmp = printBlock(inputBlock)

    if(tmp == None):
        return

    # Check if the operator of the input block to this inverter has a lower
    # precedence. If it does, then put parantheses around the input operation.
    if (not inputIsNamed and
        PRECEDENCE_LIST.index(inputBlock.getBlockType()) < PRECEDENCE_LIST.index(INVERTER_BLOCK)):
          output += "{(" + tmp + ")}"
    else:
          output += "{" + tmp + "}"

    return output

# Get a list of input blocks for a given block that are neither through IC, FALSE_in, or TRUE_in connections
# You will need this auxiliary method in this assignment
def getInputBlocks(block):
    answer = []
    for (inputb, inputs, inputn, outputn) in block.linksIN:
      if isinstance(block, TimeDelayBlock):
        if (inputb != block.IC[0]):
          answer.append(inputb)
      elif isinstance(block, TestBlock):
        if ((inputb != block.TRUE_in[0]) and (inputb != block.FALSE_in[0])):
          answer.append(inputb)
    return answer  
    
### ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ###
#
def writeToFile(string):
    filename = tkFileDialog.asksaveasfilename()
    if not filename:
       return

    # Make sure the name is valid:
    assert (filename[-4:] == ".tex"), "invalid name"

    #  writ to the file
    fd = open(filename, 'w')
    fd.write(string)
    fd.close()
    
    print ("The equations in LaTeX format have been written to " + filename)
