# Classes for the standalone simulator

import math

from CBD_DepGraph import DepGraph

from BaseBlock import *

GENERIC_BLOCK    = "Generic"
NEGATOR_BLOCK    = "Negator"
INVERTER_BLOCK   = "Inverter"
ADDER_BLOCK      = "Adder"
PRODUCT_BLOCK    = "Product"
DELAY_BLOCK      = "Delay"
INTEGRATOR_BLOCK = "Integrator"
DERIVATIVE_BLOCK = "Derivative"
CONSTANT_BLOCK   = "Constant"
TEST_BLOCK       = "Test"

INPORT_BLOCK     = "Inport"
OUTPORT_BLOCK    = "Outport"
CHILD_BLOCK      = "Child"


class ST_Simulator:
  """
  standalone simulator
  """

  def __init__(self, model, curIteration=0):
    """
    Constructor
    """
    self.__model = model
    self.__curIteration = curIteration   
    self.__depGraph = None
    self.__sortedGraph = None
    
    for block in self.__model.getBlocks():      
      if isinstance(block, ChildBlock):
        block.resetChildrenBlocksName()
                              
  def simulate(self):    
    """
    Main method
    Run simulation once
    """      
    
    if (self.__curIteration == 0):
      # we only need to flatten the model once at the beginning
      self.__flattenModel(self.__model)
      self.__removeAuxilaryModel(self.__model)
      
    if (self.__curIteration < 2):
      # we construct the dependency graph twice since it probably changed after the initial condition
      self.__depGraph = self.__createDepGraph()
      self.__sortedGraph = self.__depGraph.getStrongComponents()
      
    self.__computeBlocks()
    self.__curIteration += 1
    
  def getModel(self):
    return self.__model

  def dumpModel(self):
    """
    print out the structure of the model
    """
    blocks = self.__model.getBlocks()  
    print "*********** Model Dumping Begin ***********"
    for block in blocks:  
      block.dumpBlock()
    print "*********** Model Dumping End ***********"
 
  def dumpSignals(self, blockName = ""):
    """
    print out all signal values of a block in the model;
    if no block name is given, signals of all blocks will be dumped
    """
    blocks = self.__model.getBlocks()
    print "=========== Signals Dumping Begin ==========="
    if blockName == "":
      for block in blocks:
        block.dumpSignals()
    else:
      for block in blocks:
        if block.getBlockName() == blockName:
          block.dumpSignals()
          break
    print "=========== Signals Dumping End ==========="
          

  # Get a list of input blocks for a given block that are neither through IC, FALSE_in, or TRUE_in connections
  # You will need this auxiliary method in this assignment
  def __getInputBlocks(self, block):
      answer = []
      for (inputb, inputs, inputn, outputn) in block.linksIN:
        if isinstance(block, TimeDelayBlock):
          if (inputb != block.IC[0]):
            answer.append(inputb)
        elif isinstance(block, TestBlock):
          if ((inputb != block.TRUE_in[0]) and (inputb != block.FALSE_in[0])):
            answer.append(inputb)
      return answer   

  # Constructs the dependency graph of the current state of the CBD
  def __createDepGraph(self):      
    blocks = self.__model.getBlocks()
    depGraph = DepGraph()
    
    raise Exception, "Not Implemented Yet -- Your Job!\n"
    
    return depGraph

  # Determines if an algebraic loop describes a linear equation or not
  def __isLinear(self, strongComponent):
    
    raise Exception, "Not Implemented Yet -- Your Job!\n"
  
    return True
    
  # A helper function to check whether an input is from the strong component
  def __isInStrongComponent(self, inputBlock, strongComponent):
    for component in strongComponent:
      if (inputBlock == component):
        return True
        
    return False

  # Constructs input for a solver of systems of linear equations
  # Input consists of two matrices:
  # M1: coefficient matrix, where each row represents an equation of the system
  # M2: result matrix, where each element is the result for the corresponding equation in M1
  def __constructLinearInput(self, strongComponent):
    M1 = [] 
    M2 = []
    
    raise Exception, "Not Implemented Yet -- Your Job!\n" 
    
    return [M1, M2]
    
  # here is an example of how to compute a block
  def __computeGeneric(self, block):
    (influentBlock, s, inputn, outputn) = block.linksIN[0]
    operator = getattr(math, block.block_operator)
    output = operator(influentBlock.signals[self.__curIteration])
    block.signals.append(output)

  def __computeConstant(self, block):
    raise Exception, "Not Implemented Yet -- Your Job!\n" 

  def __computeNegator(self, block):
    raise Exception, "Not Implemented Yet -- Your Job!\n" 

  def __computeInverter(self, block):
    raise Exception, "Not Implemented Yet -- Your Job!\n" 
    
  def __computeAdder(self, block):
    raise Exception, "Not Implemented Yet -- Your Job!\n" 

  def __computeProduct(self, block):
    raise Exception, "Not Implemented Yet -- Your Job!\n" 

  def __computeDelay(self, block):
    raise Exception, "Not Implemented Yet -- Your Job!\n" 

  def __computeDerivative(self, block):
    raise Exception, "Not Implemented Yet -- Your Job!\n" 

  def __computeIntegrator(self, block):
    raise Exception, "Not Implemented Yet -- Your Job!\n" 

  def __computeTest(self, block):
    raise Exception, "Not Implemented Yet -- Your Job!\n" 
  

  def __computeInport(self, block):
    # Inport block isn't involved in simulation process
    raise ValueError("Cannot treat Inport blocks")

  def __computeOutport(self, block):
    # Outport block isn't involved in simulation process
    raise ValueError("Cannot treat Outport blocks")

  def __computeChild(self, block):
    # Child block isn't involved in simulation process
    raise ValueError("Cannot treat Child blocks")
    
  def __computeBlocks(self):
    for component in self.__sortedGraph:
      if (not self.__hasCycle(component)):
        block = component[0]    
        blockType = block.getBlockType()
        if (blockType == GENERIC_BLOCK):
          self.__computeGeneric(block)
        elif (blockType == CONSTANT_BLOCK):
          self.__computeConstant(block)
        elif (blockType == NEGATOR_BLOCK):
          self.__computeNegator(block)
        elif (blockType == INVERTER_BLOCK):
          self.__computeInverter(block)
        elif (blockType == ADDER_BLOCK):
          self.__computeAdder(block)
        elif (blockType == PRODUCT_BLOCK):
          self.__computeProduct(block)
        elif (blockType == DELAY_BLOCK):
          self.__computeDelay(block)
        elif (blockType == DERIVATIVE_BLOCK):
          self.__computeDerivative(block)
        elif (blockType == INTEGRATOR_BLOCK):
          self.__computeIntegrator(block)
        elif (blockType == TEST_BLOCK):
          self.__computeTest(block)
        elif (blockType == INPORT_BLOCK):
          self.__computeInport(block)
        elif (blockType == OUTPORT_BLOCK):
          self.__computeOutport(block)
        elif (blockType == CHILD_BLOCK):
          self.__computeChild(block)
      else:
        # Detected a strong component
        print "########### A loop is detected -> ###########"
        print component
        assert self.__isLinear(component), "Cannot solve non-linear algebraic loop"
        solverInput = self.__constructLinearInput(component)
        print "########### Input matrix for linear solver -> ###########"
        self.__printLinearInputMatrices(component, solverInput)
        self.__gaussjLinearSolver(solverInput)
        solutionVector = solverInput[1]
        print "########### Solution from linear solver -> ###########"
        for block in component:
          blockIndex = component.index(block) 
          block.signals.append(solutionVector[blockIndex])
          print block.getBlockName(), "=", solutionVector[blockIndex]

  def __checkBlock(self, block):  
          for (influentBlock, s, inputn, outputn) in block.linksIN:
            if (influentBlock.getBlockType() == CHILD_BLOCK):
              self.__flattenModel(influentBlock.model_instance)
              childrenBlocks = influentBlock.model_instance.getBlocks() 
              for childBlock in childrenBlocks:  
                if (childBlock.getBlockType() == OUTPORT_BLOCK):
                  (inBlock, s1, inputn1, outputn1) = childBlock.linksIN[0]
                  childBlockLastName = childBlock.getBlockName().split(".")[-1]
                  if (childBlockLastName == outputn):
                    block.linksIN.remove((influentBlock, s, inputn, outputn))
                    block.linksIN.append((inBlock, s1, inputn1, outputn1))
                    if (isinstance(block, TimeDelayBlock) and ((influentBlock, s, inputn, outputn) == block.IC)):
                      block.IC = (inBlock, s1, inputn1, outputn1)
                    elif isinstance(block, TestBlock):
                      if ((influentBlock, s, inputn, outputn) == block.TRUE_in):
                        block.TRUE_in = (inBlock, s1, inputn1, outputn1)
                      elif ((influentBlock, s, inputn, outputn) == block.FALSE_in):
                        block.FALSE_in = (inBlock, s1, inputn1, outputn1)
      
  def __flattenModel(self, model):
    blocks = model.getBlocks()
    for block in blocks:  
      if (len(block.linksIN) > 0):
        if (block.getBlockType() == CHILD_BLOCK):
          self.__flattenModel(block.model_instance)
          childrenBlocks = block.model_instance.getBlocks() 
          for (influentBlock, s, inputn, outputn) in block.linksIN:
            for childBlock in childrenBlocks:  
              for (inBlock, s1, inputn1, outputn1) in childBlock.linksIN:
                # the block name might be in hierarchical format (e.g., A.B.C)
                # but we only need the last part of name:
                inBlockLastName = inBlock.getBlockName().split(".")[-1]
                if ((inBlock.getBlockType() == INPORT_BLOCK) and (inBlockLastName == inputn)):
                  childBlock.linksIN.remove((inBlock, s1, inputn1, outputn1))
                  childBlock.linksIN.append((influentBlock, s, inputn, outputn))
                  if (isinstance(childBlock, TimeDelayBlock) and ((inBlock, s1, inputn1, outputn1) == childBlock.IC)):
                    childBlock.IC = (influentBlock, s, inputn, outputn)
                  elif isinstance(childBlock, TestBlock):
                    if ((inBlock, s1, inputn1, outputn1) == childBlock.TRUE_in):
                      childBlock.TRUE_in = (influentBlock, s, inputn, outputn)
                    elif ((inBlock, s1, inputn1, outputn1) == childBlock.FALSE_in):
                      childBlock.FALSE_in = (influentBlock, s, inputn, outputn)
                  self.__checkBlock(childBlock)
                  
          # remove Inport Block from child model
          for childBlock in childrenBlocks:
            for (inBlock, s1, inputn1, outputn1) in childBlock.linksIN:
              if (inBlock.getBlockType() == INPORT_BLOCK):
                childBlock.linksIN.remove((inBlock, s1, inputn1, outputn1))
                
          # remove Inport Block from child model
#          for childBlock in childrenBlocks:
#            if (childBlock.getBlockType() == INPORT_BLOCK):
#              childrenBlocks.remove(childBlock)
        else:
          self.__checkBlock(block)
          # remove Outport Block from child model
#          for (influentBlock, s, inputn, outputn) in block.linksIN:
#            if (influentBlock.getBlockType() == CHILD_BLOCK):
#              childrenBlocks = influentBlock.model_instance.getBlocks() 
#              for childBlock in childrenBlocks:
#                if (childBlock.getBlockType() == OUTPORT_BLOCK):
#                  childrenBlocks.remove(childBlock)
    childrenBlocks = []
    for block in blocks:     
      if (block.getBlockType() == CHILD_BLOCK):  
        childrenBlocks.append(block) 
    for block in childrenBlocks:
        # remove Child Block from parent model
        blocks.remove(block)
        # extend blocks of child block to parent model
        blocks.extend(block.model_instance.getBlocks())
        
  def __removeAuxilaryModel(self, model):
    blocks = model.getBlocks()
    inportBlocks = []    
    outportBlocks = []    
    for block in blocks:  
      if (len(block.linksIN) > 0):
        if (block.getBlockType() == CHILD_BLOCK):
          self.__removeAuxilaryModel(block.model_instance)
    for block in blocks:     
      if (block.getBlockType() == INPORT_BLOCK):  
        inportBlocks.append(block) 
      if (block.getBlockType() == OUTPORT_BLOCK):  
        outportBlocks.append(block)
    for block in inportBlocks:
        # remove Inport Block from parent model
        blocks.remove(block)
    for block in outportBlocks:
        # remove Outport Block from parent model
        blocks.remove(block)
    
  # Function used to detect cycles
  def __hasCycle(self, component):
    assert len(component)>=1, "A component should have at least one element"
    if len(component)>1:
      return True
    else:     
      if self.__depGraph.hasDependency(component[0], component[0]):
        return True
      else:
        return False    
    
       
  def __swap(self, a, b):
    t = a
    a = b
    b = t
    return (a, b)
  
  def __ivector(self, n):
    v = []
    for i in range(n):
      v.append(0)
    return v
    
  # Linear equation solution by Gauss-Jordan elimination
  def __gaussjLinearSolver(self, solverInput):
    M1 = solverInput[0]
    M2 = solverInput[1]
    n = len(M1)
    indxc = self.__ivector(n)
    indxr = self.__ivector(n)
    ipiv = self.__ivector(n)
    icol = 0
    irow = 0
    for i in range(n):
      big = 0.0
      for j in range(n):
        if (ipiv[j] != 1):
          for k in range(n):
            if (ipiv[k] == 0):
              if (math.fabs(M1[j][k]) >= big):
                big = math.fabs(M1[j][k])
                irow = j
                icol = k
            elif (ipiv[k] > 1): 
              raise ValueError("GAUSSJ: Singular Matrix-1")
      ipiv[icol] += 1
      if (irow != icol):
        for l in range(n):
          (M1[irow][l], M1[icol][l]) = self.__swap(M1[irow][l], M1[icol][l])
        (M2[irow], M2[icol]) = self.__swap(M2[irow], M2[icol])
      indxr[i] = irow
      indxc[i] = icol
      if (M1[icol][icol] == 0.0):
        raise ValueError("GAUSSJ: Singular Matrix-2")
      pivinv = 1.0/M1[icol][icol]
      M1[icol][icol] = 1.0
      for l in range(n):
        M1[icol][l] *= pivinv
      M2[icol] *= pivinv
      for ll in range(n):
        if (ll != icol):
          dum = M1[ll][icol]
          M1[ll][icol] = 0.0
          for l in range(n):
            M1[ll][l] -= M1[icol][l] * dum
          M2[ll] -= M2[icol] * dum
    l = n
    while (l > 0):
      l -= 1
      if (indxr[l] != indxc[l]):
        for k in range(n):
          (M1[k][indxr[l]], M1[k][indxc[l]]) = self.__swap(M1[k][indxr[l]], M1[k][indxc[l]])     
  
  # Function to print out the input matrices in a formatted way
  def __printLinearInputMatrices(self, strongComponent, solverInput):
    print "Indeces:"
    i = 0
    for block in strongComponent:
      print i, ":", block.getBlockName(), block.getBlockType()
      i += 1

    print "Matrices:"
    M1 = solverInput[0]
    M2 = solverInput[1]
    for row in M1:
      print row, "=", M2[M1.index(row)]
      
      
